-- //===========================================================================
-- //
-- // AUTHOR(S): Wishnu Prasetya, Ade Azurat
-- // PACKAGE  : (c)xMECH, class LITE, version alpha 1.3.2
-- //
-- //===========================================================================
-- //

module Lang.EXPR.Common_Parser

where

import UU.Parsing 
import UU.Scanner hiding (pKey,pSpec)
import qualified UU.Scanner (pKey,pSpec)
import PrimitiveConcepts.LC
import Configuration.Lang
import Configuration.DefaultEXPRconsts
import Lang.EXPR.PCKG
import Lang.TYPE.PCKG
import SupportLib.MoreParserBasic
import qualified SupportLib.BinExprReorder
import List(find)

pKey:: String -> Parser Token String
pSpec:: Char -> Parser Token String
pKey  = UU.Scanner.pKey
pSpec = UU.Scanner.pSpec

--
-- Data structure to contain parameters to the parsers
-- for expressions.
--

data EXPR_ParserParameter 
	=
	EXPR_ParserParameter
	{ pe_unaries :: [UserDefinedLogicalConstant],
	  --
	  -- the infixes groups should be sorted and grouped
	  -- by priority and fixity
	  --
	  pe_infixes_groups  :: [[UserDefinedLogicalConstant]],
	  pe_quantors :: [UserDefinedLogicalConstant] 
	}

type ParserIdent = Parser Token Ident
type ParserExpr  = Parser Token Expr
type ParserInfix = Parser Token (Expr->Expr->Expr)

--
-- Parsers to parse binary operators with
-- priorities and fixities.
--
-- Top interface: p_Binary
--

p_Binary parameter pArg
	=
	--
	-- mk_pPrioInfixExpr (pe_infixes_groups parameter) pArg
	--
	p_BinExpr pArg (p_dummyBinOp <|> pOp)
	
	where

	pOp = foldr1 (<|>)
              . map mk_p_binary_ 
	      . concat 
	      . pe_infixes_groups 
	      $ parameter

p_dummyBinOp :: ParserExpr
p_dummyBinOp = f <$> pKeyPosxI dummy_op_name
  where
  f pos = Ident (Identifier dummy_op_name UnknownIdentStatus TyUnknown pos)
  dummy_op_name = "!!<op>"

mk_p_binary_ :: UserDefinedLogicalConstant -> ParserExpr
mk_p_binary_ c = f <$> p_op

	where

	f pos = Ident (Identifier c_name identStatus ty pos)

	c_name = lc_name c
	p_op = mk_p_LogicalConstant c_name
	--
	-- For modularity, filling in status and type will be
	-- done by separate modules.
	--
	identStatus = UnknownIdentStatus
	ty          = TyUnknown

isLCEntryOf op (UserDefinedLogicalConstant n _ _ _) = n==op

--
-- To reorder parsed bin-expr, we need info on operators priority
-- and associativity. This info can be retrieved from the operator
-- table passed as a parameter to the parser. This will be expensive
-- since the table will be reconstructed each time we call the reordering
-- utility. Since priority and associativity are actually read from
-- a global table, and they remain unchanged for various parsing mode,
-- to improve efficiency we'll read that info directly from the 
-- infixes table from the Configuration package.
--

getPriority_ opname 
	= 
	case find (\(n,_,_,_)-> n==opname) infixes of
	
		Just (_,_,info,_) -> f info
		_		  -> error "Ack! (getPriority_)"

	where

	f (Lefty i)    = i
	f (Righty i)   = i
	f (NonAssoc i) = i

isRighty_ opname 
	= 
	case find (\(n,_,_,_)-> n==opname) infixes of
	
		Just (_,_,Righty _,_) -> True
		Just (_,_,_,_)        -> True
		_		      -> error "Ack! (isRighty_)"


instance SupportLib.BinExprReorder.Term Expr where

   isOp (Ident (Identifier n _ _ _)) 		
	=
	case find (isLCEntryOf n) default_EXPR_infixes of
        
	Just _  -> True
        _       -> False

   isOp _ = False

   prio (Ident (Identifier n _ _ _)) = getPriority_ n

   isRightAssoc (Ident (Identifier n _ _ _)) 
	=
	isRighty_ n


binTree2Expr t = case t of

	SupportLib.BinExprReorder.Node _ op u v 
		-> 
		Binary (getIdent op) (binTree2Expr u) (binTree2Expr v)

	SupportLib.BinExprReorder.Leaf a        -> a

        _             -> error "ack! (binTree2Expr)"

	where
	getIdent (Ident i) = i


p_BinExpr pArg pOp 
	=
	(\a r-> binTree2Expr . SupportLib.BinExprReorder.buildBinTree $ (a : concat r))
	<$> 
	pArg 
	<*> 
	pList ((\op a-> [op,a]) <$> pOp <*> pArg)
	

-- use pChainl to parse left associative operators
--     pChainr to parse right associative operators

pChain_x :: (Symbol c, Ord c) => Fixity -> Parser c (a->a->a) -> Parser c a -> Parser c a
pChain_x (Righty _) = pChainr_ng
pChain_x (Lefty  _) = pChainl
--
-- Non-associative will be parsed as left assoc.
-- 
pChain_x (NonAssoc _) = pChainl 

--
--for making the parser of infixes with the same priority
--
mk_pInfix_ :: [UserDefinedLogicalConstant] -> ParserInfix

mk_pInfix_ infixes 
	= 
	foldr1 (<|>) (map mk_p_binary infixes)

--
--for making the priority infix parser 
--
mk_pPrioInfixExpr :: [[UserDefinedLogicalConstant]] 
			-> ParserExpr 
			-> ParserExpr

mk_pPrioInfixExpr infixes_groups pAtom 
	= 
	foldr lipat pAtom infixes_groups
	where
	lipat infixes pLowerExpr
		=
		pChain_x assoc (mk_pInfix_ infixes) pLowerExpr
			where
			assoc = lc_fixity (head infixes)

mk_p_LogicalConstant constantname 
	=
	if is_specSym constantname     
	then pSpecPosXI (head constantname)
	else pKeyPosXI  constantname

--	then pCost 100 *> pSpecPos (head constantname)
--	else pCost 100 *> pKeyPos  constantname


mk_p_binary :: UserDefinedLogicalConstant -> ParserInfix
mk_p_binary c = f <$> p_op

	where

	f pos = Binary (Identifier c_name identStatus ty pos)

	c_name = lc_name c
	p_op = mk_p_LogicalConstant c_name
	--
	-- For modularity, filling in status and type will be
	-- done by separate modules.
	--
	identStatus = UnknownIdentStatus
	ty          = TyUnknown




--
-- Parser for Unaries.
--

mk_p_unary c = f <$> p_op

	where

	f pos = Unary (Identifier c_name identStatus ty pos)

	c_name = lc_name c
	p_op = mk_p_LogicalConstant c_name
	identStatus = UnknownIdentStatus
	ty          = TyUnknown



p_Unary :: EXPR_ParserParameter -> ParserExpr -> ParserExpr
p_Unary parameter pArg =  (foldr (<|>) pFail unaries) <*> pArg

	where

	unaries = map mk_p_unary (pe_unaries parameter)


p_UnaryString parameter pArg = f <$> pList1 pOp  <*> pArg

	where

	f ops e = foldr ($) e ops

	pOp = (foldr (<|>) pFail unaries)
	unaries = map mk_p_unary (pe_unaries parameter)

--
-- Parser for quantors
--

mk_p_quantor c = f <$> p_op

	where

	f pos = Quantor (Identifier c_name identStatus ty pos)

	c_name = lc_name c
	p_op = mk_p_LogicalConstant c_name
	identStatus = UnknownIdentStatus
	ty          = TyUnknown

p_Quantor :: EXPR_ParserParameter -> ParserIdent -> ParserExpr -> ParserExpr
p_Quantor parameter pVar pArg 
	= 
	(foldr (<|>) pFail quantors) 
	<*> pBvars 
	<*  pKeyAI ":-"
	<*> pArg 
	<*  pKeyAI "-:"
	<*> pArg

	where

	quantors = map mk_p_quantor (pe_quantors parameter)
	pBvars = pList1Sep (pSpec ',') pVar

--
-- Parser for conditional
--	
p_Cond :: ParserExpr -> ParserExpr
p_Cond pArg 
	=
	Cond
	<$>
	    pKeyPosXI "if" 
	<*> pArg
	<*  pKeyAI "then" 
	<*> pArg
	<*  pKeyAI "else" 
	<*> pArg

--
-- Parser for the SQL select expression
--
p_Select :: ParserIdent -> ParserExpr -> ParserExpr
p_Select pIdent pArg
	= 
	semfun
	<$> pKeyPosXI "select"
	<*> pArg
	<*  pKeyAI "from" 
	<*> pGenerators
	<*> (pWhere <|> pSucceed tt)

	where

	semfun pos m gens f = SQLSelect pos gens f m 	

	pGenerators = pList1Sep (pSpec ',') pGenerator

	pGenerator  = f <$>  pIdent <*> pKeyPos "<-" <*> pArg
		where
		f bvar pos gen = ListGenerator pos bvar gen

	pWhere = pKeyXI "where" *> pArg

--
-- Function application
--
p_FunApp :: ParserExpr -> ParserExpr
p_FunApp pArg = pArg <??> (lipat <$> pList1 pArg)
	where
	lipat xs f = foldl App f xs

--
-- Array or record application
--
p_ArrayRecordApp :: ParserExpr -> ParserExpr -> ParserExpr
p_ArrayRecordApp pLeftSide pArrayIndex
	=
	pLeftSide
	<??>
	(  g1 <$> pList1 pArrayElement 
	   <|> 
	   g2 <$> pList1 pRecordElement		
	)

	where

	g1 indices a = foldl lipatArray a indices
	lipatArray a (i,pos) = ArrayApp pos a i

	g2 fields  r = foldl lipatRecord r fields
	lipatRecord r (field,pos) = RecordApp pos r field

	pArrayElement  = (\pos i-> (i,pos)) 
				<$> pSpecPosXI '[' 
				<*> pArrayIndex 
				<*  pSpecAI ']'

	pRecordElement = pSpecXI '.' *> pConidPos

--
-- a range from i to j expression
--
p_FromTo :: ParserExpr -> ParserExpr -> ParserExpr
p_FromTo pArgLeft pArgRighty
	=
	semfun <$> pArgLeft
		<*> pKeyPosXI "range"
		<*  pKeyAI "from"	
		<*> pArgRighty
		<*  pKeyAI "to"	
		<*> pArgRighty
	where
	semfun a pos i j = FromTo pos a i j 

--
-- Concrete list
--
p_List :: ParserExpr -> ParserExpr
p_List pArg
	= 
	List <$> 
		    (pKeyPosXI "|" <* pSpecAI '[')	
		<*>  pListSep pComma pArg
		<*  (pSpecAI ']' <* pKeyAI "|")
--
-- Record
--
p_Record :: ParserExpr -> ParserExpr
p_Record pArg
	= 
	Record <$> 
		    pSpecPosXI '{'
		<*> pList1Sep pComma pFieldAssignment
		<*  pSpecAI '}'

	where
	pFieldAssignment = f
			<$> pConid 
			<*> pKeyPos "=" 
			<*> pArg
	f field pos e = FieldAssignment pos field e

--
-- Integer, boolean and string literals
--
p_Literal :: ParserExpr
p_Literal  = 
	Lit <$> 
	(
	      (\(i,pos)-> LitInt (read i) pos)	<$> pIntegerPos_i
	 <|>  LitBool True 			<$> pKeyPosXI "T"	
	 <|>  LitBool False			<$> pKeyPosXI "F"	
	 <|>  (\(c,pos)-> LitChar (head c) pos)	<$> pCharPosXI
	 <|>  (\(s,pos)-> LitString s pos)    	<$> pStringPosXI
	)

--
-- Parsing identifier. Note the insert cost is set to XI (maximum).
--
p_Ident :: ParserExpr
p_Ident = Ident <$> p_Identifier

p_Identifier :: ParserIdent
p_Identifier = 	f <$> pVaridPosXI
	where
	f (i,pos) = Identifier i UnknownIdentStatus TyUnknown pos

