-- //===========================================================================
-- //
-- // AUTHOR(S): Wishnu Prasetya, Ade Azurat
-- // PACKAGE  : (c)xMECH, class LITE, version alpha 1.3.2
-- //
-- //===========================================================================
-- //
--
-- This module provides utilities to reorder a list of terms
-- separated by binary operators and build a parse tree that
-- is consistent with the operators' priorities and associativities.
-- Its intended use is to speed up a parser for binary expressions.
-- (so the parser itself does not have to do the reordering, which
-- is expensive)
--

module SupportLib.BinExprReorder(
	 Term(..)
	,BinTree(..)
	,buildBinTree
	,foldBinTree
	,showBinTree
	)

where

import List(lookup)

--
-- Some useful general interface
--
class Term a where
  isOp  :: a -> Bool
  prio  :: a -> Int
  isRightAssoc :: a -> Bool 

--
-- A representation for parse trees for binary expressions
--
data BinTree a = 
	--
	-- the Int in the node is to store the priority
	-- of the operator a
	--
	  Node Int a (BinTree a) (BinTree a) 
	| Leaf a
        | EmptySlot
	deriving Show

--
-- This is the main function. It scan a list of terms
-- separated by binary operators, and build a parse
-- tree consistent with the operators' priority and
-- associativity
--
buildBinTree :: Term a => [a] -> BinTree a
buildBinTree [] = error "ack!"
buildBinTree s  = helper EmptySlot (-2) s

   where
   --
   -- q is the constructed pref-list (threaded)
   -- k is the priority of the last read symbol
   --

   helper _ _ []  = error "ack!"

   helper t _ [a] = insertR (Leaf a) t

   helper t k (a:op:s) 
--
-- the case when op binds stronger than current
--
     | k' > k    = let
                   v = Node k' op (Leaf a) EmptySlot
                   in
                   helper (insertR v t) k' s
--
-- dealing with associativity
--
     | (k' == k) && isRightAssoc op
	         = let
                   v = Node k' op (Leaf a) EmptySlot
                   in
                   helper (insertR v t) k' s

     | (k' == k) && not (isRightAssoc op)
	         = let
                    (u,v) = splitTree (k'-1) (insertR (Leaf a) t)
                    v' = Node k' op v EmptySlot
                    in
                    helper (insertR v' u) k' s
--
-- the case when op binds weaker than current
--
     | otherwise  = let
                    (u,v) = splitTree k' (insertR (Leaf a) t)
                    v' = Node k' op v EmptySlot
                    in
                    helper (insertR v' u) k' s

     where

     k' = prio op

insertR :: BinTree a -> BinTree a -> BinTree a
insertR i (Node k a u v) = Node k a u (insertR i v)
insertR i EmptySlot      = i
insertR _ (Leaf _)       = error "ack!"


splitTree :: Int -> BinTree a -> (BinTree a,BinTree a)
splitTree k' t = helper EmptySlot t
  where
  helper t t'@(Node k a u v)
     | k<=k'          = helper (insertR (Node k a u EmptySlot) t) v

     | otherwise      = (t,t')

splitTree _ _ = error "ack!"
                        

--
-- interpret a proper bintree (one that does not contain
-- empty slots)
--
foldBinTree :: Term a => (a->b,a->b->b->b) -> BinTree a -> b

foldBinTree (leaf,node) t = helper t
   where
   helper (Node _ a u v) = node a (helper u) (helper v)
   helper (Leaf a)       = leaf a
   helper _              = error "ack!"

--
-- simple show utility for BinTree
--
showBinTree :: Term a => (a->String) -> BinTree a -> String
showBinTree showTerm t = foldBinTree (leaf,node) t []
	where
        leaf a     = (showTerm a ++)
	node a t u = ("(" ++) 
			. t 
			. (" " ++) 
			. (showTerm a ++)
			. (" " ++)
			. u 
			. (")" ++)



-- ================================================================
-- TEST
-- ================================================================

instance Term Char where
  isOp a = a `elem` "+-*/#%"
  prio a = case a of
	'+'	->	0
	'-'	->	0
	'*'	->	1
	'/'	->	1
	'#'	->	2
	'%'	->	3
	_	-> error "ack! (prio)"

  isRightAssoc a = a `elem` "+*#"

  


